##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Basic ZPublisher request management.

    + monkeypatch for break traversing without explicit acquisition

"""

import logging
from Acquisition import aq_base
from Acquisition.interfaces import IAcquirer
from Products.CMFCore.interfaces import IContentish
from Products.CMFPlone.interfaces import IPloneSiteRoot
from ZPublisher.interfaces import UseTraversalDefault
from ZPublisher.BaseRequest import typeCheck
from zExceptions import Forbidden
from zExceptions import NotFound
from zope.component import queryMultiAdapter
from zope.interface import Interface

from experimental.noacquisition import config


logger = logging.getLogger('experimental.noacquisition')


def publishTraverse(self, request, name):
    object = self.context
    URL=request['URL']

    if name[:1]=='_':
        raise Forbidden("Object name begins with an underscore at: %s" % URL)

    subobject = UseTraversalDefault  # indicator
    try:
        if hasattr(object,'__bobo_traverse__'):
            try:
                subobject=object.__bobo_traverse__(request, name)
                if type(subobject) is type(()) and len(subobject) > 1:
                    # Add additional parents into the path
                    # XXX There are no tests for this:
                    request['PARENTS'][-1:] = list(subobject[:-1])
                    object, subobject = subobject[-2:]
            except (AttributeError, KeyError, NotFound), e:
                # Try to find a view
                subobject = queryMultiAdapter((object, request), Interface, name)
                if subobject is not None:
                    # OFS.Application.__bobo_traverse__ calls
                    # REQUEST.RESPONSE.notFoundError which sets the HTTP
                    # status code to 404
                    request.response.setStatus(200)
                    # We don't need to do the docstring security check
                    # for views, so lets skip it and return the object here.
                    if IAcquirer.providedBy(subobject):
                        subobject = subobject.__of__(object)
                    return subobject
                # No view found. Reraise the error raised by __bobo_traverse__
                raise e
    except UseTraversalDefault:
        pass
    if subobject is UseTraversalDefault:
        # No __bobo_traverse__ or default traversal requested
        # Try with an unacquired attribute:
        if hasattr(aq_base(object), name):
            subobject = getattr(object, name)
        else:
            # We try to fall back to a view:
            subobject = queryMultiAdapter((object, request), Interface,
                                          name)
            if subobject is not None:
                if IAcquirer.providedBy(subobject):
                    subobject = subobject.__of__(object)
                return subobject

            # And lastly, of there is no view, try acquired attributes, but
            # only if there is no __bobo_traverse__:
            try:
                subobject=getattr(object, name)
                if subobject:
                    logger.debug(
                        'traverse without explicit acquisition '
                        'object=%r name=%r subobject=%r url=%r referer=%r',
                        object, name, subobject,
                        request.get('ACTUAL_URL'),
                        request.get('HTTP_REFERER', '-')
                    )
                    ##
                    ## STOP TRAVERSING WITHOUT EXPLICIT ACQUISITION
                    ##
                    if (IContentish.providedBy(subobject) or
                            IPloneSiteRoot.providedBy(subobject)):
                        logger.warning(
                            'traverse without explicit acquisition '
                            'object=%r name=%r subobject=%r url=%r referer=%r',
                            object, name, subobject,
                            request.get('ACTUAL_URL'),
                            request.get('HTTP_REFERER', '-')
                        )
                        if not config.DRYRUN:
                            subobject = None
                            raise AttributeError

                # Again, clear any error status created by __bobo_traverse__
                # because we actually found something:
                request.response.setStatus(200)
            except AttributeError:
                pass

            # Lastly we try with key access:
            if subobject is None:
                try:
                    subobject = object[name]
                except TypeError: # unsubscriptable
                    raise KeyError(name)

    # Ensure that the object has a docstring, or that the parent
    # object has a pseudo-docstring for the object. Objects that
    # have an empty or missing docstring are not published.
    doc = getattr(subobject, '__doc__', None)
    if not doc:
        raise Forbidden(
            "The object at %s has an empty or missing " \
            "docstring. Objects must have a docstring to be " \
            "published." % URL
            )

    # Hack for security: in Python 2.2.2, most built-in types
    # gained docstrings that they didn't have before. That caused
    # certain mutable types (dicts, lists) to become publishable
    # when they shouldn't be. The following check makes sure that
    # the right thing happens in both 2.2.2+ and earlier versions.

    if not typeCheck(subobject):
        raise Forbidden(
            "The object at %s is not publishable." % URL
            )

    return subobject



import logging
from Products.Archetypes.utils import shasattr
from Products.CMFCore.interfaces import IContentish
from Products.CMFPlone.interfaces import IPloneSiteRoot
from ZPublisher import xmlrpc
from webdav.NullResource import NullResource
from zope.component import queryMultiAdapter
from zope.interface import Interface
from experimental.noacquisition import config


logger = logging.getLogger('experimental.noacquisition')


def __bobo_traverse__(self, REQUEST, name):
    """Allows transparent access to session subobjects.
    """
    # sometimes, the request doesn't have a response, e.g. when
    # PageTemplates traverse through the object path, they pass in
    # a phony request (a dict).

    RESPONSE = getattr(REQUEST, 'RESPONSE', None)

    # Is it a registered sub object
    data = self.getSubObject(name, REQUEST, RESPONSE)
    if data is not None:
        return data
    # Or a standard attribute (maybe acquired...)
    target = None
    method = REQUEST.get('REQUEST_METHOD', 'GET').upper()
    # Logic from "ZPublisher.BaseRequest.BaseRequest.traverse"
    # to check whether this is a browser request
    if (len(REQUEST.get('TraversalRequestNameStack', ())) == 0 and
        not (method in ('GET', 'HEAD', 'POST') and not
             isinstance(RESPONSE, xmlrpc.Response))):
        if shasattr(self, name):
            target = getattr(self, name)
    else:
        if shasattr(self, name):  # attributes of self come first
            target = getattr(self, name)
        else:  # then views
            target = queryMultiAdapter((self, REQUEST), Interface, name)
            if target is not None:
                # We don't return the view, we raise an
                # AttributeError instead (below)
                target = None
            else:  # then acquired attributes
                target = getattr(self, name, None)
                if target is not None:
                    logger.debug(
                        'traverse without explicit acquisition '
                        'object=%r name=%r subobject=%r url=%r referer=%r',
                        self, name, target,
                        REQUEST.get('ACTUAL_URL'),
                        REQUEST.get('HTTP_REFERER', '-')
                    )
                    ##
                    ## STOP TRAVERSING WITHOUT EXPLICIT ACQUISITION
                    ##
                    if REQUEST.get('ACTUAL_URL') and (
                            IContentish.providedBy(target) or
                            IPloneSiteRoot.providedBy(target)):
                        logger.warning(
                            'traverse without explicit acquisition '
                            'object=%r name=%r subobject=%r url=%r referer=%r',
                            self, name, target,
                            REQUEST.get('ACTUAL_URL'),
                            REQUEST.get('HTTP_REFERER', '-')
                        )
                        if not config.DRYRUN:
                            target = None

    if target is not None:
        return target
    elif (method not in ('GET', 'POST') and not
          isinstance(RESPONSE, xmlrpc.Response) and
          REQUEST.maybe_webdav_client):
        return NullResource(self, name, REQUEST).__of__(self)
    else:
        # Raising AttributeError will look up views for us
        raise AttributeError(name)



# mostly grabbed from
# https://github.com/plone/Products.CMFPlone/blob/publication-through-explicit-acquisition/Products/CMFPlone/tests/test_bad_acquisition.py

import pkg_resources
import os.path
import unittest
from urllib2 import HTTPError
from plone.testing.z2 import Browser
from plone.app.testing import TEST_USER_NAME, TEST_USER_PASSWORD

try:
    pkg_resources.get_distribution('plone.app.contenttypes')
except pkg_resources.DistributionNotFound:
    HAS_PACONTENTTYPES = False
else:
    HAS_PACONTENTTYPES = True

from experimental.noacquisition.testing import BASE_FUNCTIONAL_TESTING
from experimental.noacquisition import config


def dummy_image():
    filename = os.path.join(os.path.dirname(__file__), u'image.gif')
    if HAS_PACONTENTTYPES:
        from plone.namedfile.file import NamedBlobImage
        return NamedBlobImage(
            data=open(filename, 'r').read(),
            filename=filename
        )
    else:
        return open(filename, 'r').read()


class TestBadAcquisition(unittest.TestCase):

    layer = BASE_FUNCTIONAL_TESTING

    def setUp(self):
        self.default_dryrun_config = config.DRYRUN
        config.DRYRUN = False
        self.portal = self.layer['portal']
        self.app = self.layer['app']
        self.portal.invokeFactory('Document', 'a_page')
        self.assertTrue('a_page' in self.portal.objectIds())
        self.portal.invokeFactory('Folder', 'a_folder')
        self.assertTrue('a_folder' in self.portal.objectIds())
        self.portal.invokeFactory('Image', id='a_image', image=dummy_image())
        self.assertTrue('a_image' in self.portal.objectIds())
        import transaction
        transaction.commit()
        self.browser = Browser(self.app)
        self.browser.addHeader(
            'Authorization',
            'Basic %s:%s' % (TEST_USER_NAME, TEST_USER_PASSWORD,)
        )

    def tearDown(self):
        config.DRYRUN = self.default_dryrun_config

    def test_not_found_when_acquired_content(self):
        self.browser.open(self.portal.a_page.absolute_url())
        error = None
        try:
            url = self.portal.absolute_url() + '/a_folder/a_page'
            self.browser.open(url)
        except HTTPError, e:
            error = e
        self.assertTrue(
            error is not None,
            msg='Acquired content should not be published.')
        self.assertEqual(404, error.code)

    def test_not_found_when_acquired_image_traverser(self):
        url = self.portal.a_image.absolute_url() + "/@@images/image"
        self.browser.open(url)
        error = None
        try:
            url = self.portal.absolute_url() + \
                '/a_folder/a_image/@@images/image'
            self.browser.open(url)
        except HTTPError, e:
            error = e
        self.assertTrue(
            error is not None,
            msg='Acquired content should not be published.')
        self.assertEqual(404, error.code)

    # This currently fails, because p.a.testeing create de plonesite with
    # "plone" id but there is also a view with same name ....
    # def test_not_found_when_acquired_siteroot_1(self):
    #     error = None
    #     try:
    #         url = self.portal.absolute_url() + "/" + self.portal.getId()
    #         self.browser.open(url)
    #     except HTTPError, e:
    #         error = e
    #     self.assertTrue(
    #         error is not None,
    #         msg='Acquired content should not be published.')
    #     self.assertEqual(404, error.code)

    # This currently fails, because p.a.testeing create de plonesite with
    # "plone" id but there is also a view with same name ....
    # def test_not_found_when_acquired_siteroot_2(self):
    #     error = None
    #     try:
    #         url = self.portal.a_folder.absolute_url() + "/" + self.portal.getId()
    #         self.browser.open(url)
    #     except HTTPError, e:
    #         error = e
    #     self.assertTrue(
    #         error is not None,
    #         msg='Acquired content should not be published.')
    #     self.assertEqual(404, error.code)

    # TODO: unfinshed
    # def test_traverse_vhm(self):
    #     url = "http://nohost/VirtualHostBase/http/example.org:80/plone" + \
    #         "/VirtualHostRoot/a_page"
    #     self.browser.open(url)

    def test_traverse_portal_skin_object(self):
        url = self.portal.absolute_url() + '/logo.png'
        self.browser.open(url)
        url = self.portal.a_page.absolute_url() + '/logo.png'
        self.browser.open(url)
        url = self.portal.a_image.absolute_url() + '/logo.png'
        self.browser.open(url)



from plone.app.testing.layers import FunctionalTesting
from Products.CMFPlone.testing import ProductsCMFPloneLayer


class BaseLayer(ProductsCMFPloneLayer):
    def setUpZope(self, app, configurationContext):  # pylint: disable=W0613
        # import plone.app.imaging
        # self.loadZCML(package=plone.app.imaging, name='configure.zcml')
        import experimental.noacquisition
        # self.loadZCML(package=experimental.noacquisition, name='configure.zcml')
        self.loadZCML(package=experimental.noacquisition, name='testing.zcml')


BASE = BaseLayer()


BASE_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(BASE, ),
    name="experimental.noacquisiton:Functional"
)



from setuptools import setup, find_packages


version = '1.0.0b1'

setup(name='experimental.noacquisition',
      version=version,
      description="No acquistion during publish traverse",
      long_description=(open("README.rst").read() + "\n" +
                        open("CHANGES.rst").read()),
      # Get more strings from
      # http://pypi.python.org/pypi?:action=list_classifiers
      classifiers=[
          "Framework :: Plone",
          "Framework :: Zope2",
          "Programming Language :: Python",
      ],
      keywords='monkeypatch traverse',
      author='Mauro Amico',
      author_email='mauro.amico@gmail.com',
      url='http://pypi.org/pypi/collective/experimental.noacquisition',
      license='BSD',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      namespace_packages=['experimental', ],
      include_package_data=True,
      zip_safe=False,
      test_suite="experimental.noacquisition",
      install_requires=[
          'setuptools',
          # -*- Extra requirements: -*-
          'collective.monkeypatcher',
          'Zope2>=2.13.4,<=2.13.23'  # paranoid
      ],
      extras_require={
          'test': ['Products.CMFPlone[test]', 'Products.PloneTestCase']
      },
      entry_points="""
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )
